from getpass import getpass

import click
from oic import rndstr
from oic.oic import Client
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

import unikube.cli.console as console
from unikube import settings
from unikube.helpers import compare_current_and_latest_versions


@click.command()
@click.option("--email", "-e", type=str, help="Authentication email")
@click.option("--password", "-p", type=str, help="Authentication password")
@click.pass_obj
def login(ctx, email, password, **kwargs):
    """
    Authenticate with a Unikube host. The default login process is a Browser-based method.
    If you want to login without being redirected to the Browser, you can just specify the parameter
    ``-e`` for email and enable the direct login method. For a non-interactive login, you can provide
    ``-p`` along with the password.
    """
    compare_current_and_latest_versions()
    if email or password:
        if not email:
            email = click.prompt("email", type=str)
        if not password:
            password = getpass("password:")
        return password_flow(ctx, email, password)
    return web_flow(ctx)


def password_flow(ctx, email, password):
    response = ctx.auth.login(
        email,
        password,
    )
    if response["success"]:
        try:
            token = ctx.auth.token_from_response(response)
        except Exception as e:
            console.debug(e)
            console.debug(response)
            console.error("Login failed. Your token does not match.")
            return False

        if token["given_name"]:
            console.success(f'Login successful. Hello {token["given_name"]}!')
        else:
            console.success("Login successful.")
    else:
        console.error("Login failed. Please check email and password.")
    return True


def web_flow(ctx):
    client = Client(client_authn_method=CLIENT_AUTHN_METHOD)
    issuer = f"{settings.AUTH_DEFAULT_HOST}/auth/realms/unikube"
    client.provider_config(issuer)

    state = rndstr()
    nonce = rndstr()

    # 1. run callback server
    from unikube.authentication.web import run_callback_server

    port = run_callback_server(state, nonce, client, ctx)

    # 2. send to login with redirect url.
    args = {
        "client_id": "cli",
        "response_type": ["token"],
        "response_mode": "form_post",
        "scope": ["openid"],
        "nonce": nonce,
        "state": state,
        "redirect_uri": f"http://localhost:{port}",
    }

    auth_req = client.construct_AuthorizationRequest(request_args=args)
    login_url = auth_req.request(client.authorization_endpoint)
    console.info("If your Browser does not open automatically, go to the following URL and login:")
    console.link(login_url)
    click.launch(login_url)
    return True


@click.command()
@click.pass_obj
def logout(ctx, **kwargs):
    """
    Log out of a Unikube host.
    """

    ctx.auth.logout()
    console.info("Logout completed.")

    return True


@click.command()
@click.option("--token", "-t", is_flag=True, default=False, help="Show token information.")
@click.pass_obj
def status(ctx, token=False, **kwargs):
    """
    View authentication status.
    """

    response = ctx.auth.verify()

    # show token information
    if token:
        console.info(f"access token: {ctx.auth.general_data.authentication.access_token}")
        console.echo("---")
        console.info(f"refresh token: {ctx.auth.general_data.authentication.refresh_token}")
        console.echo("---")
        console.info(f"requesting party token: {ctx.auth.general_data.authentication.requesting_party_token}")
        console.echo("")

    if response["success"]:
        console.success("Authentication verified.")
    else:
        console.info("Authentication could not be verified.")

    return True



from unikube.storage.local_storage import LocalStorage
from unikube.storage.types import GeneralData


class LocalStorageGeneral(LocalStorage):
    table_name = GeneralData().id
    pydantic_class = GeneralData

    document_id = GeneralData().id

    def get(self) -> GeneralData:
        data = super().get(id=self.document_id)
        return self.pydantic_class(**data.dict())

    def set(self, data: GeneralData) -> None:
        super().set(id=self.document_id, data=data)

    def delete(self) -> None:
        super().delete(id=self.document_id)



# heavily inspired by yaspin (https://github.com/pavdmyt/yaspin/blob/master/yaspin/core.py)
import sys
from functools import wraps
from itertools import cycle
from threading import Event, Lock, Thread
from typing import Callable, Dict


class Spinner(object):
    def __init__(self, text=""):
        self.text = text
        self.start()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.thread.is_alive():
            self.stop()
        return False

    def start(self):
        self.thread = Thread(target=self._spin)
        self._stdout_lock = Lock()
        self.stop_event = Event()
        self.thread.start()

    def stop(self):
        if self.thread:
            self.stop_event.set()
            self.thread.join()
        sys.stdout.write("\r")
        self._clear_line()

    @staticmethod
    def _clear_line():
        sys.stdout.write("\033[K")

    def message(self, message):
        with self._stdout_lock:
            sys.stdout.write("\r")
            self._clear_line()
            sys.stdout.write(f"{message}\n")

    def success(self, message):
        self.message(f"\033[92m✔\033[0m {message}")

    def info(self, message):
        self.message(f"\033[96mℹ\033[0m {message}")

    def error(self, message):
        self.message(f"\033[91m✘\033[0m {message}")

    def change_spinner_text(self, text):
        self.text = text

    def _spin(self):
        frames = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
        interval = 80 * 0.001
        _cycle = cycle(frames)
        while not self.stop_event.is_set():
            char = next(_cycle)
            with self._stdout_lock:
                sys.stdout.write("\r")
                sys.stdout.write(f"{char} {self.text}")
                self._clear_line()
                sys.stdout.flush()
            self.stop_event.wait(interval)


def spinner(spin_message: str, success_message: str, error_messages: Dict[Callable[..., Exception], str] = None):
    """
    Any Exception type provided in `error_messages` will be handled with the according message being displayed
    as an error message. All other exceptions will bubble.

    Usage:

    @spinner("Doing xyz ...", "xyz done!", {ValueError: "Could not do xyz because of bad input"})
    def xyz(*args, **kwargs):
        # do unspeakable things
    """
    error_messages = error_messages or {}

    def decorator(fnc):
        @wraps(fnc)
        def wrapper(*args, **kwargs):
            with Spinner(spin_message) as spinner:
                try:
                    rval = fnc(*args, **kwargs)
                except Exception as e:
                    for exception_type in type(e).mro():
                        if exception_type in error_messages:
                            spinner.error(error_messages[exception_type])
                            break
                    else:
                        raise e
                else:
                    spinner.success(success_message)
                    return rval

        return wrapper

    return decorator



from pydantic import BaseModel

from unikube.authentication.types import AuthenticationData
from unikube.context.types import ContextData
from unikube.types import ConfigurationData


class TinyDatabaseData(BaseModel):
    id: str


class GeneralData(TinyDatabaseData):
    id: str = "general"
    authentication: AuthenticationData = AuthenticationData()


class UserData(TinyDatabaseData):
    context: ContextData = ContextData()
    config: ConfigurationData = ConfigurationData()



import os
import socket
import sys
import tempfile
from collections import OrderedDict
from typing import List, Tuple

import click
import click_spinner

from unikube import settings
from unikube.cli import console
from unikube.cli.helper import age_from_timestamp
from unikube.graphql_utils import GraphQL
from unikube.local.providers.helper import get_cluster_or_exit
from unikube.local.system import Docker, KubeAPI, KubeCtl, Telepresence
from unikube.settings import UNIKUBE_FILE
from unikube.unikubefile.selector import unikube_file_selector


def _is_local_port_free(port):
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if a_socket.connect_ex(("127.0.0.1", int(port))) == 0:
        return False
    else:
        return True


def get_deck_from_arguments(ctx, organization_id: str, project_id: str, deck_id: str):
    # context
    organization_id, project_id, deck_id = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization_id, project_argument=project_id, deck_argument=deck_id
    )

    # argument
    if not deck_id:
        deck_id = console.deck_list(ctx, organization_id=organization_id, project_id=project_id)
        if not deck_id:
            exit(1)

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    id
                    title
                    environment {
                        namespace
                    }
                    project {
                        id
                    }
                }
            }
            """,
            query_variables={"id": deck_id},
        )
        deck = data["deck"]
        project_id = deck["project"]["id"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # cluster data
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    if project_id not in [cluster.id for cluster in cluster_list]:
        console.info(f"The project cluster for '{project_id}' is not up or does not exist yet.", _exit=True)

    cluster_data = ctx.cluster_manager.get(id=project_id)
    if not cluster_data:
        console.error("The cluster could not be found.", _exit=True)

    return cluster_data, deck


def argument_apps(k8s, apps: List[str], multiselect: bool = False) -> List[str]:
    if not apps:
        app_choices = [
            pod.metadata.name
            for pod in k8s.get_pods().items
            if pod.status.phase not in ["Terminating", "Evicted", "Pending"]
        ]
        message = "Please select an app" if not multiselect else "Please select one or multiple apps"
        kwargs = {
            "message": message,
            "choices": app_choices,
            "multiselect": multiselect,
        }
        if multiselect:
            kwargs["transformer"] = lambda result: f"{', '.join(result)}"
            apps = console.list(**kwargs)
        else:
            apps = [console.list(**kwargs)]

    if not apps:
        console.error("No apps available.", _exit=True)

    if apps and any(c_app not in [pod.metadata.name for pod in k8s.get_pods().items] for c_app in apps):
        console.error("Some apps do not exist.", _exit=True)

    return apps


def argument_app(k8s, app: str) -> str:
    return argument_apps(k8s, [app] if app else [])[0]


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def list(ctx, organization, project, deck, **kwargs):
    """List all apps."""

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # list
    k8s = KubeAPI(provider_data, deck)
    pod_table = []

    def _ready_ind(c) -> Tuple[bool, str]:
        # get container count
        if c is None:
            container_count = 0
            ready_count = 0
        else:
            container_count = len(c)
            ready_count = sum([val.ready for val in c])
        return container_count == ready_count, f"{ready_count}/{container_count}"

    for pod in k8s.get_pods().items:
        if pod.status.phase in ["Terminating", "Evicted", "Pending"]:
            continue
        all_ready, count = _ready_ind(pod.status.container_statuses)
        pod_table.append(
            OrderedDict(
                {
                    "name": pod.metadata.name,
                    "ready": count,
                    "state": "Ok" if all_ready else "Not Ok",
                    "age": age_from_timestamp(pod.metadata.creation_timestamp.timestamp()),
                }
            )
        )
    console.table(
        data=pod_table,
        headers={
            "name": "Name",
            "ready": "Ready",
            "state": "State",
            "age": "Age",
        },
    )


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def info(ctx, app, organization, project, deck, **kwargs):
    """Display the status for the given app name."""

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # shell
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    # get the data of the selected pod
    data = k8s.get_pod(app)
    pod_status = data.status

    console.info(f"This app runs {len(pod_status.container_statuses)} container(s).")
    for idx, status in enumerate(pod_status.container_statuses):
        console.info(f"Container {idx + 1}: {status.image}")
        print("\nStartup command from workload manifest:")
        console.table(
            [
                ("Command", " ".join(data.spec.containers[idx].command) if data.spec.containers[idx].command else None),
                ("Args", " ".join(data.spec.containers[idx].args) if data.spec.containers[idx].args else None),
            ]
        )
        print("\nApp status:")
        console.table(
            [
                {"State": "Running", "Value": status.state.running.started_at if status.state.running else None},
                {
                    "State": "Terminated",
                    "Value": status.state.terminated.finished_at if status.state.terminated else None,
                },
                {"State": "Waiting", "Value": status.state.waiting.message if status.state.waiting else None},
            ]
        )

    conditions = []
    for condition in pod_status.conditions:
        conditions.append(
            OrderedDict(
                {
                    "type": condition.type,
                    "status": condition.status,
                    "reason": condition.reason,
                    "last_transition_time": condition.last_transition_time,
                    "last_probe_time": condition.last_probe_time,
                    "message": condition.message,
                }
            )
        )

    if conditions:
        conditions = sorted(conditions, key=lambda x: x.get("last_transition_time").timestamp())
        # print a line for padding on the console
        print()
        console.info("All conditions for this app:")
        console.table(
            conditions,
            headers={
                "type": "Type",
                "status": "Status",
                "reason": "Reason",
                "last_transition_time": "Time",
                "last_probe_time": "Probe Time",
                "message": "Message",
            },
        )
    else:
        console.info("No condition to display")


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.option("--container", "-c", help="Specify the container in this app")
@click.pass_obj
def shell(ctx, app, organization=None, project=None, deck=None, container=None, **kwargs):
    """
    Drop into an interactive shell.
    """

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # shell
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    # get the data of the selected pod
    data = k8s.get_pod(app)
    telepresence = Telepresence(provider_data)

    # the corresponding deployment by getting rid of the pod name suffix
    deployment = "-".join(data.metadata.name.split("-")[0:-2])

    # 1. check if this pod is of a switched deployment (in case of an active Telepresence)
    if telepresence.is_swapped(deployment, namespace=data.metadata.namespace):
        # the container name generated in "app switch" for that pod
        container_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
            project=cluster_data.name.lower(), deck=deck["title"].lower(), name=deployment.lower()
        ).replace(":", "")

        if Docker().check_running(container_name):
            # 2. Connect to that container
            # 2.a connect using Docker
            Docker().exec(container_name, "/bin/sh", interactive=True)
        else:
            console.error(
                "This is a Telepresence Pod with no corresponding Docker container "
                "running in order to connect (inconsistent state?)"
            )

    else:
        if not container and len(data.spec.containers) > 1:
            container = console.container_list(data=data)
            if not container:
                return None

        # 2.b connect using kubernetes
        KubeCtl(provider_data).exec_pod(
            app, deck["environment"][0]["namespace"], "/bin/sh", interactive=True, container=container
        )


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_context
def exec(ctx, **kwargs):
    ctx.forward(shell)


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.option("--deployment", help="Specify the deployment if not set in the Unikubefile")
@click.option("--unikubefile", help="Specify the path to the Unikubefile", type=str)
@click.option(
    "--no-build", "-n", is_flag=True, help="Do not build a new container image for the switch operation", default=False
)
@click.pass_obj
def switch(
    ctx, app, organization, project, deck, deployment, unikubefile: str = None, no_build: bool = False, **kwargs
):
    """
    Switch a running deployment with a local Docker container.
    """

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)

    # unikube file input
    try:
        unikube_file = unikube_file_selector.get(path_unikube_file=unikubefile)
        unikube_file_app = unikube_file.get_app(name=app)
    except Exception as e:
        console.debug(e)
        console.error("Invalid 'app' argument.", _exit=True)

    # 2: Get a deployment
    # 2.1.a Check the deployment identifier
    if not deployment and unikube_file_app:
        # 1.1.b check the unikubefile
        deployment = unikube_file_app.get_deployment()
        if not deployment:
            console.error("Please specify the 'deployment' key of your app in your unikube.yaml.", _exit=True)
    else:
        console.error(
            "Please specify the deployment either using the '--deployment' option or in the unikube.yaml. "
            "Run 'unikube app switch' in a directory containing the unikube.yaml file.",
            _exit=True,
        )

    # 2.2 Fetch available "deployment:", deployments
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    deployments(level: "local") {
                        id
                        title
                        description
                        ports
                        isSwitchable
                    }
                    environment {
                        id
                        type
                        valuesPath
                        namespace
                    }
                }
            }
            """,
            query_variables={
                "id": deck["id"],
            },
        )
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    target_deployment = None
    for _deployment in data["deck"]["deployments"]:
        if _deployment["title"] == deployment:
            target_deployment = _deployment

    # 2.3 Check and select deployment data
    if target_deployment is None:
        console.error(
            f"The deployment '{deployment}' you specified could not be found.",
            _exit=True,
        )

    ports = target_deployment["ports"].split(",")
    deployment = target_deployment["title"]
    namespace = deck["environment"][0]["namespace"]

    console.info("Please wait while unikube prepares the switch.")
    with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
        # check telepresence
        provider_data = cluster.storage.get()
        telepresence = Telepresence(provider_data)

        available_deployments = telepresence.list(namespace, flat=True)
        if deployment not in available_deployments:
            console.error(
                "The given deployment cannot be switched. " f"You may have to run 'unikube deck install {deck}' first.",
                _exit=True,
            )

        is_swapped = telepresence.is_swapped(deployment, namespace)

        k8s = KubeAPI(provider_data, deck)
        # service account token, service cert
        service_account_tokens = k8s.get_serviceaccount_tokens(deployment)

    # 3: Build an new Docker image
    # 3.1 Grab the docker file
    context, dockerfile, target = unikube_file_app.get_docker_build()
    if not target:
        target = ""
    console.debug(f"{context}, {dockerfile}, {target}")

    # 3.2 Set an image name
    image_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
        project=cluster_data.name.replace(" ", "").lower(), deck=deck["title"], name=deployment
    )

    docker = Docker()

    if is_swapped:
        console.warning("It seems this app is already switched in another process. ")
        if click.confirm("Do you want to kill it and switch here?"):
            telepresence.leave(deployment, namespace, silent=True)
            if docker.check_running(image_name):
                docker.kill(name=image_name)
        else:
            sys.exit(0)

    # 3.3 Build image
    if not docker.image_exists(image_name) or not no_build:
        if no_build:
            console.warning(f"Ignoring --no-build since the required image '{image_name}' does not exist")
        console.info(f"Building a Docker image for {dockerfile} with context {context}")
        with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
            status, msg = docker.build(image_name, context, dockerfile, target)
        if not status:
            console.debug(msg)
            console.error("Failed to build Docker image.", _exit=True)

        console.info(f"Docker image successfully built: {image_name}")

    # 4. Start the Telepresence session
    # 4.1 Set the right intercept port
    port = unikube_file_app.get_port()
    if port is None:
        port = str(ports[0])
        if len(ports) > 1:
            console.warning(
                f"No port specified although there are multiple ports available: {ports}. "
                f"Defaulting to port {port} which might not be correct."
            )
    if port not in ports:
        console.error(f"The specified port {port} is not in the rage of available options: {ports}", _exit=True)
    if not _is_local_port_free(port):
        console.error(
            f"The local port {port} is busy. Please stop the application running on " f"this port and try again.",
            _exit=True,
        )

    # 4.2 See if there are volume mounts
    mounts = unikube_file_app.get_mounts()
    console.debug(f"Volumes requested: {mounts}")
    # mount service tokens
    if service_account_tokens:
        tmp_sa_token = tempfile.NamedTemporaryFile(delete=True)
        tmp_sa_cert = tempfile.NamedTemporaryFile(delete=True)
        tmp_sa_token.write(service_account_tokens[0].encode())
        tmp_sa_cert.write(service_account_tokens[1].encode())
        tmp_sa_token.flush()
        tmp_sa_cert.flush()
        mounts.append((tmp_sa_token.name, settings.SERVICE_TOKEN_FILENAME))
        mounts.append((tmp_sa_cert.name, settings.SERVICE_CERT_FILENAME))
    else:
        tmp_sa_token = None
        tmp_sa_cert = None

    # 4.3 See if there special env variables
    envs = unikube_file_app.get_environment()
    console.debug(f"Envs requested: {envs}")

    # 4.4 See if there is a run command to be executed
    command = unikube_file_app.get_command(port=port)
    console.debug(f"Run command: {command}")

    console.info("Starting your container, this may take a while to become effective")

    telepresence.swap(deployment, image_name, command, namespace, envs, mounts, port)
    if docker.check_running(image_name):
        docker.kill(name=image_name)
    if tmp_sa_token:
        tmp_sa_token.close()
        tmp_sa_cert.close()


@click.command()
@click.argument("app", required=False)
@click.option("--container", "-c", help="Specify the container in this app")
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.option("--follow", "-f", is_flag=True, default=False, help="Follow the log stream.")
@click.pass_obj
def logs(ctx, app, container=None, organization=None, project=None, deck=None, follow=False, **kwargs):
    """
    Display the logs for an app. If this app contains multiple containers, specify the ``container``
    argument or choose it from the interactive selector. You can follow the log stream if you specify the
    ``-f`` flag.
    """

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # log
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    # get the data of the selected pod
    if not container:
        data = k8s.get_pod(app)
        if len(data.spec.containers) > 1:
            container = console.container_list(data=data)
            if not container:
                return None

    logs = k8s.get_logs(app, follow, container=container)

    # output
    click.echo(logs)


@click.command()
@click.argument("app", required=False)
@click.option("--init", "-i", is_flag=True, help="Display environment variables for the init container", default=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def env(ctx, app, init, organization, project, deck, **kwargs):
    """
    Display the environment variables for the given app. This prints the environment variables for all containers. You
    can print the environment variables for all init containers with the ``-i`` flag.
    """

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # env
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    # get the data of the selected pod
    data = k8s.get_pod(app)

    if init:
        containers = data.spec.init_containers
    else:
        containers = data.spec.containers

    if containers:

        console.info(f"This app runs {len(containers)} container(s).")
        for idx, container in enumerate(containers):
            console.info(f"Container {idx + 1}: {container.image}")
            env_vars = []

            def _value_from(s) -> Tuple[str, str]:
                # return an indicator if this values comes from a secret and the name
                if s.config_map_key_ref:
                    return "ConfigMap", s.config_map_key_ref
                elif s.field_ref and s.field_ref.field_path:
                    return "Field", s.field_ref.field_path
                elif s.resource_field_ref:
                    return "ResourceField", s.resource_field_ref
                elif s.secret_key_ref and s.secret_key_ref.name:
                    return "Secret", f"Secret: {s.secret_key_ref.name} Key: {s.secret_key_ref.key}"

            if container.env:
                for env in container.env:
                    if env.value_from:
                        type, source = _value_from(env.value_from)
                    else:
                        type, source = "Definition", "-"
                    env_vars.append(
                        OrderedDict(
                            {
                                "name": env.name,
                                "value": str(env.value[:50]) + "..."
                                if env.value and len(env.value) > 50
                                else env.value,
                                "source_type": type,
                                "path": source,
                            }
                        )
                    )
            if container.env_from:
                for env in container.env_from:
                    if env.config_map_ref:
                        _cm = k8s.get_configmap(env.config_map_ref.name)
                        if _cm.data:
                            for k, v in _cm.data.items():
                                env_vars.append(
                                    OrderedDict(
                                        {
                                            "name": k,
                                            "value": str(v[:50]) + "..." if v and len(v) > 50 else v,
                                            "source_type": "ConfigMap",
                                            "path": _cm.metadata.name,
                                        }
                                    )
                                )

            console.table(
                env_vars,
                headers={
                    "name": "Name",
                    "value": "Value",
                    "path": "Path",
                    "source_type": "Source Type",
                },
            )
    else:
        console.info("No container running")


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def update(ctx, app, organization, project, deck, **kwargs):
    """
    Trigger a forced update of the given app. This command creates a new app instance.
    """

    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # delete pod
    k8s = KubeAPI(provider_data, deck)
    apps = argument_apps(k8s, [app] if app else [], multiselect=True)
    [k8s.delete_pod(app) for app in apps]
    console.info(f"The app(s) {', '.join(apps)} are currently updating and do not exist anymore.")



from typing import List, Union

import unikube.cli.console as console
from unikube.cli.console.input import get_identifier_or_pass
from unikube.context.helper import convert_project_argument_to_uuid
from unikube.graphql_utils import GraphQL


def project_list(
    ctx, organization_id: str = None, filter: List[str] = None, excludes: List[str] = None
) -> Union[None, str]:
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID) {
                allProjects(organizationId: $organization_id) {
                    results {
                        title
                        id
                        organization {
                            id
                            title
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": organization_id,
            },
        )
        project_list = data["allProjects"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    selection = console.list(
        message="Please select a project",
        choices=[project["title"] for project in project_list],
        identifiers=[project["id"] for project in project_list],
        filter=filter,
        excludes=excludes,
        help_texts=[project["organization"]["title"] for project in project_list],
        message_no_choices="No projects available!",
    )
    if selection is None:
        return None

    # get identifier if available
    project_argument = get_identifier_or_pass(selection)

    project_id = convert_project_argument_to_uuid(
        ctx.auth, argument_value=project_argument, organization_id=organization_id
    )
    return project_id



from typing import Tuple
from uuid import UUID

from slugify import slugify

from unikube.cli import console
from unikube.graphql_utils import GraphQL


class ArgumentError(Exception):
    pass


# uuid validation
def is_valid_uuid4(uuid: str):
    try:
        _ = UUID(uuid, version=4)
        return True
    except Exception:
        return False


# context arguments
def __select_result(argument_value: str, results: list, exception_message: str = "context"):
    # slugify
    if slugify(argument_value) != argument_value:
        title_list = [item["title"] for item in results]
    else:
        title_list = [slugify(item["title"]) for item in results]

    # check if name/title exists and is unique
    count = title_list.count(argument_value)
    if count == 0:
        raise ArgumentError(f"{exception_message.capitalize()} name/slug does not exist.")

    if count > 1:
        raise ArgumentError(f"{exception_message.capitalize()} name/slug is not unique.")

    # find index
    try:
        index = title_list.index(argument_value)
    except Exception:
        raise ArgumentError(f"Invalid {exception_message} name/slug.")

    # convert name/title to uuid
    return results[index]["id"]


def convert_organization_argument_to_uuid(auth, argument_value: str) -> str:
    # uuid provided (no conversion required)
    if is_valid_uuid4(argument_value):
        return argument_value

    # get available context options or use provided data (e.g. from previous query)
    graph_ql = GraphQL(authentication=auth)
    data = graph_ql.query(
        """
        query {
            allOrganizations {
                results {
                    title
                    id
                }
            }
        }
        """
    )

    results = data["allOrganizations"]["results"]
    return __select_result(argument_value, results, exception_message="organization")


def convert_project_argument_to_uuid(auth, argument_value: str, organization_id: str = None) -> str:
    # uuid provided (no conversion required)
    if is_valid_uuid4(argument_value):
        return argument_value

    # get available context options or use provided data (e.g. from previous query)
    graph_ql = GraphQL(authentication=auth)
    data = graph_ql.query(
        """
        query($organization_id: UUID) {
            allProjects(organizationId: $organization_id) {
                results {
                    title
                    id
                }
            }
        }
        """,
        query_variables={
            "organization_id": organization_id,
        },
    )

    results = data["allProjects"]["results"]
    return __select_result(argument_value, results, exception_message="project")


def convert_deck_argument_to_uuid(
    auth, argument_value: str, organization_id: str = None, project_id: str = None
) -> str:
    # uuid provided (no conversion required)
    if is_valid_uuid4(argument_value):
        return argument_value

    # get available context options or use provided data (e.g. from previous query)
    graph_ql = GraphQL(authentication=auth)
    data = graph_ql.query(
        """
        query($organization_id: UUID, $project_id: UUID) {
            allDecks(organizationId: $organization_id, projectId: $project_id) {
                results {
                    title
                    id
                }
            }
        }
        """,
        query_variables={
            "organization_id": organization_id,
            "project_id": project_id,
        },
    )

    results = data["allDecks"]["results"]
    return __select_result(argument_value, results, exception_message="deck")


def convert_context_arguments(
    auth, organization_argument: str = None, project_argument: str = None, deck_argument: str = None
) -> Tuple[str, str, str]:
    try:
        # organization
        if organization_argument:
            organization_id = convert_organization_argument_to_uuid(auth, organization_argument)
        else:
            organization_id = None

        # project
        if project_argument:
            project_id = convert_project_argument_to_uuid(auth, project_argument, organization_id=organization_id)
        else:
            project_id = None

        # deck
        if deck_argument:
            deck_id = convert_deck_argument_to_uuid(
                auth, deck_argument, organization_id=organization_id, project_id=project_id
            )
        else:
            deck_id = None
    except Exception as e:
        console.error(e, _exit=True)

    return organization_id, project_id, deck_id



# -*- coding: utf-8 -*-
import os
from typing import List, Optional, Tuple

from pydantic import BaseModel

from unikube.context.types import ContextData
from unikube.unikubefile.unikube_file import UnikubeFile


class UnikubeFileContext(BaseModel):
    organization: Optional[str] = None
    project: Optional[str] = None
    deck: Optional[str] = None


class UnikubeFileBuild(BaseModel):
    context: str
    dockerfile: Optional[str] = "Dockerfile"
    target: Optional[str] = None


class UnikubeFileApp(BaseModel):
    unikube_file: str
    name: str
    build: UnikubeFileBuild
    deployment: str
    port: Optional[int] = None
    command: str
    volumes: Optional[List[str]] = None
    env: Optional[List[dict]] = None

    def get_docker_build(self) -> Tuple[str, str, str]:
        if self.build:
            base_path = os.path.dirname(self.unikube_file)
            path = os.path.abspath(os.path.join(base_path, self.build.context))
            dockerfile = os.path.join(base_path, self.build.dockerfile)
            target = self.build.target
            return path, dockerfile, target
        else:
            return os.path.abspath("."), "Dockerfile", ""

    def get_command(self, **format) -> Optional[str]:
        if not self.command:
            return None

        command = self.command.format(**format)
        return command.split(" ")

    def get_port(self) -> Optional[str]:
        if self.port:
            return str(self.port)
        else:
            return None

    def get_deployment(self) -> Optional[str]:
        if self.deployment:
            return str(self.deployment)
        else:
            return None

    def get_mounts(self) -> List[Tuple[str, str]]:
        mounts = []
        if self.volumes:
            base_path = os.path.dirname(self.unikube_file)
            for mount in self.volumes:
                mount = mount.split(":")
                source = os.path.abspath(os.path.join(base_path, mount[0]))
                target = mount[1]
                mounts.append((source, target))
        return mounts

    def get_environment(self) -> List[Tuple[str, str]]:
        envs = []
        if self.env:
            for env in self.env:
                envs.append(list(env.items())[0])
        return envs


class UnikubeFile_1_0(UnikubeFile, BaseModel):
    version: Optional[str]
    context: Optional[UnikubeFileContext] = None
    apps: List[UnikubeFileApp]

    def get_context(self) -> ContextData:
        if not self.context:
            return ContextData()

        return ContextData(
            organization_id=self.context.organization,
            project_id=self.context.project,
            deck_id=self.context.deck,
        )

    def get_app(self, name: str = None) -> UnikubeFileApp:
        # default name
        if not name:
            name = "default"

        for app in self.apps:
            if app.name == name:
                return app

        if name != "default":
            raise ValueError("Invalid name.")

        return self.apps[0]



import click
from tabulate import tabulate

from unikube import settings


def table(data, headers={}):
    click.echo(
        tabulate(
            data,
            headers=headers,
            tablefmt=settings.CLI_TABLEFMT,
        )
    )



from tests.login_testcase import LoginTestCase
from unikube.cli import deck
from unikube.commands import ClickContext


class DeckTestCase(LoginTestCase):
    def test_deck_info(self):
        result = self.runner.invoke(
            deck.info,
            ["buzzword-counter"],
            obj=ClickContext(),
        )

        self.assertIn("Key", result.output)
        self.assertIn("Value", result.output)
        self.assertIn("buzzword-counter", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_info_not_existing_deck(self):

        result = self.runner.invoke(
            deck.info,
            ["not_existing_deck"],
            obj=ClickContext(),
        )
        self.assertIn("[ERROR] Deck name/slug does not exist.\n", result.output)

    def test_deck_list(self):

        result = self.runner.invoke(
            deck.list,
            obj=ClickContext(),
        )

        self.assertIn("project", result.output)
        self.assertIn("id", result.output)
        self.assertIn("title", result.output)
        self.assertIn("buzzword-counter", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_deck_ingress(self):
        result = self.runner.invoke(
            deck.ingress,
            ["4634368f-1751-40ae-9cd7-738fcb656a0d"],
            obj=ClickContext(),
        )

        self.assertIn(
            "[ERROR] The project cluster does not exist. Please be sure to run 'unikube project up' first.\n",
            result.output,
        )
        self.assertEqual(result.exit_code, 1)



import pytest
from requests import HTTPError, Session

from unikube.helpers import (
    check_environment_type_local_or_exit,
    download_manifest,
    download_specs,
    environment_type_from_string,
    get_requests_session,
)


def test_get_requests_session():
    access_token = ""
    assert type(access_token) is str

    session = get_requests_session(access_token=access_token)
    assert type(session) is Session


def test_download_specs():
    environment_id = "WRONG"
    access_token = ""

    with pytest.raises(HTTPError) as pytest_wrapped_e:
        _ = download_specs(access_token=access_token, environment_id=environment_id)

    assert pytest_wrapped_e.type == HTTPError


def test_download_manifest():
    deck = {
        "environment": [
            {
                "id": None,
            }
        ],
        "project": {
            "id": None,
        },
    }

    class Authentication:
        def refresh(self):
            return {"success": True, "response": {"access_token": ""}}

    access_token = ""
    authentication = Authentication()

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        _ = download_manifest(deck=deck, authentication=authentication, access_token=access_token)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1


def test_environment_type_from_string():
    result = environment_type_from_string("")
    assert result is None


def test_check_environment_type_local_or_exit():
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        deck = {"environment": {0: {"type": "REMOTE"}}}
        check_environment_type_local_or_exit(deck)
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1



from .container import container_list
from .deck import deck_list
from .exit import exit_generic_error, exit_login_required
from .input import confirm, list
from .logger import debug, echo, error, info, link, success, warning
from .orga import organization_list
from .output import table
from .project import project_list



from tests.login_testcase import LoginTestCase
from unikube.cli import orga, project
from unikube.commands import ClickContext


class ProjectTestCase(LoginTestCase):
    def test_project_info(self):
        result = self.runner.invoke(
            project.info,
            ["buzzword-counter"],
            obj=ClickContext(),
        )

        self.assertIn("Key", result.output)
        self.assertIn("Value", result.output)
        self.assertIn("buzzword-counter", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_info_not_existing_project(self):
        result = self.runner.invoke(
            project.info,
            ["not-existing-project"],
            obj=ClickContext(),
        )

        self.assertIn("[ERROR] Project name/slug does not exist.\n", result.output)

    def test_project_list(self):
        result = self.runner.invoke(
            project.list,
            obj=ClickContext(),
        )

        self.assertIn("id", result.output)
        self.assertIn("name", result.output)
        self.assertIn("buzzword-counter", result.output)
        self.assertEqual(result.exit_code, 0)



from typing import Union

import unikube.cli.console as console
from unikube.cli.console.input import get_identifier_or_pass
from unikube.context.helper import convert_deck_argument_to_uuid
from unikube.graphql_utils import GraphQL


def deck_list(ctx, organization_id: str = None, project_id: str = None) -> Union[None, str]:
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID, $project_id: UUID) {
                allDecks(organizationId: $organization_id, projectId: $project_id) {
                    results {
                        title
                        id
                        project {
                            id
                            title
                            organization {
                                id
                            }
                        }
                    }
                }
            }
            """,
            query_variables={
                "organization_id": organization_id,
                "project_id": project_id,
            },
        )
        deck_list = data["allDecks"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    selection = console.list(
        message="Please select a deck",
        choices=[deck["title"] for deck in deck_list],
        identifiers=[deck["id"] for deck in deck_list],
        help_texts=[deck["project"]["title"] for deck in deck_list],
        message_no_choices="No decks available!",
    )
    if selection is None:
        return None

    # get identifier if available
    deck_argument = get_identifier_or_pass(selection)

    deck_id = convert_deck_argument_to_uuid(
        ctx.auth, argument_value=deck_argument, organization_id=organization_id, project_id=project_id
    )
    return deck_id



class ClickContext(object):
    def __init__(self):
        from unikube.authentication.authentication import get_authentication
        from unikube.context.context import Context
        from unikube.local.providers.manager import K8sClusterManager
        from unikube.storage.general import LocalStorageGeneral

        self.auth = get_authentication()
        self.storage_general = LocalStorageGeneral()
        self.context: Context = Context(auth=self.auth)
        self.cluster_manager = K8sClusterManager()



from typing import Optional

from unikube.cli import console
from unikube.graphql_utils import GraphQL


def organization_id_2_display_name(ctx, id: str = None) -> str:
    if not id:
        return "-"

    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID!) {
                organization(id: $id) {
                    title
                }
            }
            """,
            query_variables={
                "id": id,
            },
        )
        title = data["organization"]["title"]
    except Exception as e:
        console.debug(e)
        title = "-"

    return f"{title} ({id})"


def project_id_2_display_name(ctx, id: str = None) -> Optional[str]:
    if not id:
        return "-"

    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID!) {
                project(id: $id) {
                    title
                }
            }
            """,
            query_variables={
                "id": id,
            },
        )
        title = data["project"]["title"]
    except Exception as e:
        console.debug(e)
        title = "-"

    return f"{title} ({id})"


def deck_id_2_display_name(ctx, id: str = None) -> Optional[str]:
    if not id:
        return "-"

    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID!) {
                deck(id: $id) {
                    title
                }
            }
            """,
            query_variables={
                "id": id,
            },
        )
        title = data["deck"]["title"]
    except Exception as e:
        console.debug(e)
        title = "-"

    return f"{title} ({id})"



import os
from typing import Union

import click
import yaml

from unikube.settings import UNIKUBE_FILE
from unikube.unikubefile.unikube_file import (
    UnikubeFile,
    UnikubeFileError,
    UnikubeFileNotFoundError,
    UnikubeFileVersionError,
)
from unikube.unikubefile.unikube_file_1_0 import UnikubeFile_1_0


class UnikubeFileSelector:
    def __init__(self, options: dict):
        self.options = options

    @staticmethod
    def __convert_apps_to_list(path_unikube_file: str, data: dict):
        apps_list = []
        for name, item in data["apps"].items():
            item["unikube_file"] = path_unikube_file
            item["name"] = name
            apps_list.append(item)
        data["apps"] = apps_list
        return data

    def get(self, path_unikube_file: str = None) -> Union[UnikubeFile, UnikubeFile_1_0, None]:
        # default file path
        if not path_unikube_file:
            path_unikube_file = os.path.join(os.getcwd(), UNIKUBE_FILE)

        # load unikube file + get version
        try:
            with click.open_file(path_unikube_file) as unikube_file:
                data = yaml.load(unikube_file, Loader=yaml.FullLoader)
        except FileNotFoundError:
            raise UnikubeFileNotFoundError

        # add & format data
        try:
            data = self.__convert_apps_to_list(path_unikube_file=path_unikube_file, data=data)
        except Exception:
            raise UnikubeFileError("Invalid unikube file.")

        # version
        version = str(data.get("version", "latest"))
        data["version"] = version

        # get class
        unikube_file_class = self.options.get(version)
        if not unikube_file_class:
            raise UnikubeFileVersionError

        return unikube_file_class(**data)


unikube_file_selector = UnikubeFileSelector(
    options={
        "latest": UnikubeFile_1_0,
        "1": UnikubeFile_1_0,
        "1.0": UnikubeFile_1_0,
    }
)



import re
from typing import Any, Callable, List, Union

from InquirerPy import inquirer

import unikube.cli.console as console


def get_identifier_or_pass(selection: str) -> str:
    # get identifier if available
    # example: "PROJECT_NAME (IDENTIFIER)"

    identifier_search = re.search("(?<=\\()[^)]*(?=\\))", selection)
    try:
        project_argument = identifier_search.group(0)
    except Exception:
        project_argument = selection

    return project_argument


def resolve_duplicates(
    choices: List[str],
    identifiers: List[str],
    help_texts: Union[List[str], None] = None,
    help_texts_always: bool = False,
) -> List[str]:
    # detect duplicates
    duplicates_mask = [True if choices.count(choice) > 1 else False for choice in choices]

    # add identifiers to duplicates
    choices_resolved = []
    for choice, identifier, duplicate in zip(choices, identifiers, duplicates_mask):
        if duplicate:
            choices_resolved.append(f"{choice} ({identifier})")
        else:
            choices_resolved.append(choice)

    # help texts
    def add_help_text(choice, help_text):
        return f"{choice} - {help_text}"

    choices_resolved_with_help_text = []
    if help_texts:
        for choice, help_text, duplicate in zip(choices_resolved, help_texts, duplicates_mask):
            # add help_text always
            if help_texts_always and help_text:
                choices_resolved_with_help_text.append(add_help_text(choice, help_text))
                continue

            # add help_text for duplicates only
            if duplicate and help_text:
                choices_resolved_with_help_text.append(add_help_text(choice, help_text))
            else:
                choices_resolved_with_help_text.append(choice)

        choices_resolved = choices_resolved_with_help_text

    return choices_resolved


def filter_by_identifiers(choices: List[str], identifiers: List[str], filter: Union[List[str], None]) -> List[str]:
    if filter is None:
        return choices

    choices_filtered = []
    for choice, identifier in zip(choices, identifiers):
        if any(f in choice for f in filter) or identifier in filter:
            choices_filtered.append(choice)
    return choices_filtered


def exclude_by_identifiers(choices: List[str], identifiers: List[str], excludes: Union[List[str], None]) -> List[str]:
    if not excludes:
        return choices

    choices_excluded = []
    for choice, identifier in zip(choices, identifiers):
        if any(exclude in choice for exclude in excludes) or identifier in excludes:
            continue
        choices_excluded.append(choice)
    return choices_excluded


# input
def list(
    message: str,
    choices: List[str],
    identifiers: Union[List[str], None] = None,
    filter: Union[List[str], None] = None,
    excludes: Union[List[str], None] = None,
    help_texts: Union[List[str], None] = None,
    allow_duplicates: bool = False,
    message_no_choices: str = "No choices available!",
    multiselect: bool = False,
    transformer: Callable[[Any], str] = None,
) -> Union[None, List[str]]:
    # choices exist
    if not len(choices) > 0:
        console.info(message_no_choices)
        return None

    # handle duplicates
    if not allow_duplicates:
        if identifiers:
            choices_duplicates = resolve_duplicates(choices=choices, identifiers=identifiers, help_texts=help_texts)
        else:
            choices_duplicates = set(choices)
    else:
        choices_duplicates = choices

    # filter
    choices_filtered = filter_by_identifiers(choices=choices_duplicates, identifiers=identifiers, filter=filter)

    # exclude
    choices_excluded = exclude_by_identifiers(choices=choices_filtered, identifiers=identifiers, excludes=excludes)

    # prompt
    answer = inquirer.fuzzy(
        message=message,
        choices=choices_excluded,
        multiselect=multiselect,
        transformer=transformer,
        keybindings={"toggle": [{"key": "space"}]},
    ).execute()
    if not answer:
        return None

    return answer


def confirm(
    question: str = "Do want to continue? [N/y]: ",
    values: List[str] = ["y", "Y", "yes", "Yes"],
) -> bool:
    # confirm action by user input

    confirm = input(question)
    if confirm not in values:
        return False
    return True



import datetime
import errno
import os
import socket
from typing import List


def port_in_use(port: int):
    """
    Checks whether a port is available on the local system.
    """
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    location = ("127.0.0.1", port)
    result_of_check = a_socket.connect_ex(location)

    return result_of_check == 0


def check_ports(port_list: List[int]):
    """
    Takes a list of integers and check whether those ports are available.

    Returns a list of ports which are busy.
    """
    return list(filter(port_in_use, port_list))


def exist_or_create(folder):
    if not os.path.exists(os.path.dirname(folder)):
        try:
            os.makedirs(os.path.dirname(folder))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise


def age_from_timestamp(timestamp):
    delta = datetime.timedelta(seconds=datetime.datetime.now(tz=datetime.timezone.utc).timestamp() - timestamp)
    if delta >= datetime.timedelta(days=1):
        result = f"{delta.days}d"
    elif delta >= datetime.timedelta(hours=1):
        result = f"{delta.seconds // 3600}h"
    elif delta >= datetime.timedelta(minutes=10):
        result = f"{delta.seconds // 60}m"

    elif delta >= datetime.timedelta(minutes=1):
        result = f"{delta.seconds // 60}m{delta.seconds % 60}s"
    else:
        result = f"{delta.seconds}s"
    return result



import sys
from time import sleep, time

import click
import click_spinner

import unikube.cli.console as console
from unikube import settings
from unikube.cli.console.helpers import project_id_2_display_name
from unikube.cli.helper import check_ports
from unikube.graphql_utils import GraphQL
from unikube.helpers import check_running_cluster
from unikube.local.providers.types import K8sProviderType
from unikube.local.system import Docker, KubeAPI, Telepresence


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.pass_obj
def list(ctx, organization, **kwargs):
    """
    Display a table of all available project names alongside with the ids.
    """

    _ = ctx.auth.refresh()

    # context
    organization_id, _, _ = ctx.context.get_context_ids_from_arguments(organization_argument=organization)

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($organization_id: UUID) {
                allProjects(organizationId: $organization_id) {
                    results {
                        title
                        id
                        description
                    }
                }
            }
            """,
            query_variables={"organization_id": organization_id},
        )
        project_list = data["allProjects"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # console
    if len(project_list) < 1:
        console.info("No projects available. Please go to https://app.unikube.io and create a project.", _exit=True)

    console.table(
        data={
            "id": [p["id"] for p in project_list],
            "title": [p["title"] for p in project_list],
            "description": [p["description"] for p in project_list],
        },
        headers=["id", "name", "description"],
    )


@click.command()
@click.argument("project", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.pass_obj
def info(ctx, project=None, organization=None, **kwargs):
    """
    Displays the id, title and optional description of the selected project.
    """

    _ = ctx.auth.refresh()

    # context
    organization_id, project_id, _ = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project
    )

    # argument
    if not project_id:
        project_id = console.project_list(ctx, organization_id=organization_id)
        if not project_id:
            return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID!) {
                project(id: $id) {
                    id
                    title
                    description
                    specRepository
                    specRepositoryBranch
                    organization {
                        title
                    }
                }
            }
            """,
            query_variables={"id": project_id},
        )
        project_selected = data["project"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # console
    if project_selected:
        project_selected["organization"] = project_selected.pop("organization").get("title", "-")
        project_selected["repository"] = project_selected.pop("specRepository")
        project_selected["repository branch"] = project_selected.pop("specRepositoryBranch")

        console.table(
            data={
                "key": [k for k in project_selected.keys()],
                "value": [v for v in project_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Project does not exist.")


@click.command()
@click.argument("project", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--ingress", help="Overwrite the ingress port for the project from cluster settings", default=None)
@click.option(
    "--provider",
    "-p",
    help="Specify the Kubernetes provider type for this cluster (default uses k3d)",
    default=settings.UNIKUBE_DEFAULT_PROVIDER_TYPE.name,
)
@click.option("--workers", help="Specify count of k3d worker nodes", default=1)
@click.pass_obj
def up(ctx, project=None, organization=None, ingress=None, provider=None, workers=None, **kwargs):
    """
    This command starts or resumes a Kubernetes cluster for the specified project. As it is a selection command, the
    project can be specified and/or filtered in several ways:

    * as a positional argument, id or project title can be specified, or from a set context
    * as an interactive selection from available projects
    * via ``-o`` or ``--organization`` option, specifying an organisation to which a project belongs

    """

    _ = ctx.auth.refresh()

    if not Docker().daemon_active():
        console.error("Docker is not running. Please start Docker before starting a project.", _exit=True)

    # context
    organization_id, project_id, _ = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project
    )

    # cluster information
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    cluster_id_list = [item.id for item in cluster_list]

    # argument
    if not project_id:
        project_id = console.project_list(ctx, organization_id=organization_id, excludes=cluster_id_list)
        if not project_id:
            return None

    if project_id in cluster_id_list:
        console.info(f"Project '{project_id_2_display_name(ctx=ctx, id=project_id)}' is already up.", _exit=True)

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                project(id: $id) {
                    title
                    id
                    organization {
                        id
                    }
                    clusterSettings {
                        id
                        port
                    }
                    organization {
                        title
                    }
                }
            }
            """,
            query_variables={
                "id": project_id,
            },
        )
        project_selected = data["project"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    if not project_selected:
        console.info(
            f"The project '{project_id_2_display_name(ctx=ctx, id=project_id)}' could not be found.", _exit=True
        )

    try:
        cluster_provider_type = K8sProviderType[provider]
    except KeyError:
        console.error(
            f"The provider '{provider}' is not supported. Please use "
            f"one of: {','.join(opt.name for opt in K8sProviderType)}",
            _exit=True,
        )

    check_running_cluster(ctx, cluster_provider_type, project_selected)

    # get project id
    if ingress is None:
        ingress = project_selected["clusterSettings"]["port"]

    if not_available_ports := check_ports([ingress]):
        console.error(
            "Following ports are currently busy, however needed to spin up the cluster: {}".format(
                ", ".join([str(port) for port in not_available_ports])
            ),
            _exit=True,
        )

    # cluster up
    cluster_data = ctx.cluster_manager.get(id=project_selected["id"])
    cluster_data.name = project_selected["title"]
    ctx.cluster_manager.set(id=project_selected["id"], data=cluster_data)

    cluster = ctx.cluster_manager.select(cluster_data=cluster_data, cluster_provider_type=cluster_provider_type)
    console.info(
        f"Setting up a Kubernetes cluster (with provider {provider}) for " f"project '{cluster.display_name}'."
    )

    if not cluster.exists():
        console.info(f"Kubernetes cluster for '{cluster.display_name}' does not exist, creating it now.")
        with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
            success = cluster.create(
                ingress_port=ingress,
                workers=workers,
            )

    # start
    else:
        console.info(f"Kubernetes cluster for '{cluster.display_name}' already exists, starting it now.")
        with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
            success = cluster.start()

    # console
    if success:
        console.info("Now connecting Telepresence daemon. You probably have to enter your 'sudo' password.")
        provider_data = cluster.storage.get()
        k8s = KubeAPI(provider_data)
        timeout = time() + 60  # wait one minute
        while not k8s.is_available or time() > timeout:
            sleep(1)
        if not k8s.is_available:
            console.error(
                "There was an error bringing up the project cluster. The API was not available within the"
                "expiration period.",
                _exit=True,
            )
        Telepresence(cluster.storage.get()).start()
        console.success("The project cluster is up.")
    else:
        console.error("The project cluster could not be started.")


@click.command()
@click.argument("project", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.pass_obj
def down(ctx, project=None, organization=None, **kwargs):
    """
    Stop/pause cluster.
    """

    # context
    organization_id, project_id, _ = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project
    )

    # cluster
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)

    # argument
    if not project_id:
        project_id = console.project_list(
            ctx, organization_id=organization_id, filter=[cluster.id for cluster in cluster_list]
        )
        if not project_id:
            return None

    # check if project is in local storage
    if project_id not in [cluster.id for cluster in cluster_list]:
        console.info(
            f"The project cluster for '{project_id_2_display_name(ctx=ctx, id=project_id)}' is not up or does not exist yet.",
            _exit=True,
        )

    # get cluster
    cluster = None
    for cluster_data in cluster_list:
        if cluster_data.id == project_id:
            cluster = ctx.cluster_manager.select(
                cluster_data=cluster_data,
            )
            break

    # cluster down
    if not cluster.exists():
        # something went wrong or cluster was already delete from somewhere else
        console.info(f"No Kubernetes cluster to stop for '{cluster.display_name}'", _exit=True)

    if not cluster.ready():
        console.info(f"Kubernetes cluster for '{cluster.display_name}' is not running", _exit=True)

    console.info("Stopping Telepresence daemon.")
    Telepresence(cluster.storage.get()).stop()

    # stop cluster
    console.info(f"Stopping Kubernetes cluster for '{cluster.display_name}'")
    success = cluster.stop()

    # console
    if success:
        console.success("The project cluster is down.")
    else:
        console.error("The cluster could not be stopped.")


@click.command()
@click.argument("project", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.pass_obj
def delete(ctx, project=None, organization=None, **kwargs):
    """
    Delete the current project and all related data.
    """

    # context
    organization_id, project_id, _ = ctx.context.get_context_ids_from_arguments(
        organization_argument=organization, project_argument=project
    )

    # cluster
    cluster_list = ctx.cluster_manager.get_cluster_list()

    # argument
    if not project_id:
        project_id = console.project_list(
            ctx, organization_id=organization_id, filter=[cluster.id for cluster in cluster_list]
        )
        if not project_id:
            return None

    if project_id not in [cluster.id for cluster in cluster_list]:
        console.info(
            f"The project cluster for '{project_id_2_display_name(ctx=ctx, id=project_id)}' could not be found.",
            _exit=True,
        )

    # initial warning
    console.warning("Deleting a project will remove the cluster including all of its data.")

    # confirm question
    confirm = input("Do want to continue [N/y]: ")
    if confirm not in ["y", "Y", "yes", "Yes"]:
        console.info("No action taken.", _exit=True)

    # get cluster
    cluster = None
    for cluster_data in cluster_list:
        if cluster_data.id == project_id:
            cluster = ctx.cluster_manager.select(
                cluster_data=cluster_data,
            )
            break

    # delete cluster
    if not cluster.exists():
        ctx.cluster_manager.delete(cluster.id)
        console.info(f"No Kubernetes cluster to delete for '{cluster.display_name}', nothing to do.", _exit=True)

    success = cluster.delete()

    # console
    if success:
        console.success("The project was deleted successfully.")
        ctx.cluster_manager.delete(cluster.id)
    else:
        console.error("The cluster could not be deleted.")


@click.command()
@click.pass_obj
def prune(ctx, **kwargs):
    """
    Remove unused clusters.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query {
                allProjects {
                    results {
                        id
                    }
                }
            }
            """
        )
        projects = data["allProjects"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    # cluster
    cluster_list = ctx.cluster_manager.get_cluster_list()

    # select clusters to prune
    prune_clusters = []
    for cluster_data in cluster_list:
        if cluster_data.id not in [project["id"] for project in projects]:
            prune_clusters.append(cluster_data)

    for cluster_data in prune_clusters:
        console.info(f"It seems like the project for cluster '{cluster_data.name}' has been deleted.")

        # confirm question
        confirmed = console.confirm(question="Do want to remove the cluster? [N/y]: ")
        if not confirmed:
            console.info("No action taken.")
            continue

        # delete
        try:
            cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
            success = cluster.delete()
            if success:
                console.success("The project was deleted successfully.")
                ctx.cluster_manager.delete(cluster.id)
        except Exception as e:
            console.debug(e)
            console.error("The cluster could not be deleted.")



import os

from setuptools import find_packages, setup

base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, "VERSION")) as f:
    VERSION = f.read()


with open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

DESCRIPTION = "This is the unikube.io command line interface"


setup(
    name="unikube",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=VERSION,
    py_modules=["unikube"],
    install_requires=[
        "click==8.0.4",
        "click-spinner==0.1.10",
        "colorama~=0.4.4",
        "inquirerpy>=0.2.4,<0.4.0",
        "tabulate~=0.8.9",
        "pydantic>=1.7.3,<1.10.0",
        "pyyaml>=5.4",
        "tinydb>=3.15.2,<4.8.0",
        "requests>=2.25.1,<2.28.0",
        "pyjwt[crypto]>=2.0.1,<2.4.0",
        "gql>=3.0,<3.1",
        "semantic-version>=2.8.4,<2.10.0",
        "kubernetes>=11.0.0",
        "retrying~=1.3.3",
        "oic==1.3.0",
        "python-slugify>=5.0.2,<6.1.0",
        "click-didyoumean~=0.3.0",
        "requests-toolbelt~=0.9.1",
    ],
    python_requires="~=3.7",
    packages=find_packages(),
    url="https://github.com/unikubehq/cli",
    project_urls={
        "Source": "https://github.com/unikubehq/cli",
        "Documentation": "https://cli.unikube.io",
        "Bug Tracker": "https://github.com/unikubehq/cli/issues",
    },
    author="Michael Schilonka",
    author_email="michael@blueshoe.de",
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={"console_scripts": ["unikube=unikube.commands:cli"]},
)



from typing import Union

import unikube.cli.console as console
from unikube.cli.console.input import get_identifier_or_pass
from unikube.context.helper import convert_organization_argument_to_uuid
from unikube.graphql_utils import GraphQL


def organization_list(ctx) -> Union[None, str]:
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query {
                allOrganizations {
                    results {
                        id
                        title
                    }
                }
            }
            """
        )
        organization_list = data["allOrganizations"]["results"]
    except Exception as e:
        console.debug(e)
        console.exit_generic_error()

    selection = console.list(
        message="Please select an organization",
        choices=[organization["title"] for organization in organization_list],
        identifiers=[organization["id"] for organization in organization_list],
        message_no_choices="No organizations available!",
    )
    if selection is None:
        return None

    # get identifier if available
    organization_argument = get_identifier_or_pass(selection)

    organization_id = convert_organization_argument_to_uuid(ctx.auth, argument_value=organization_argument)
    return organization_id



from typing import Optional

from pydantic import BaseModel


class ContextData(BaseModel):
    organization_id: Optional[str] = None
    project_id: Optional[str] = None
    deck_id: Optional[str] = None



from unikube.storage.general import LocalStorageGeneral
from unikube.storage.local_storage import LocalStorage
from unikube.storage.types import UserData


class LocalStorageUser(LocalStorage):
    table_name = "user"
    pydantic_class = UserData

    def __init__(self, user_email) -> None:
        super().__init__()

        self.user_email = user_email

    def get(self) -> UserData:
        try:
            data = super().get(id=self.user_email)
            return self.pydantic_class(**data.dict())
        except Exception:
            return UserData()

    def set(self, data: UserData) -> None:
        super().set(id=self.user_email, data=data)

    def delete(self) -> None:
        super().delete(id=self.user_email)


def get_local_storage_user():
    try:
        local_storage_general = LocalStorageGeneral()
        general_data = local_storage_general.get()
        local_storage_user = LocalStorageUser(user_email=general_data.authentication.email)
        return local_storage_user
    except Exception:
        return None



import os
from abc import ABC, abstractmethod
from typing import List, Tuple, Union

from unikube import settings
from unikube.cli import console
from unikube.context.helper import convert_context_arguments, is_valid_uuid4
from unikube.context.types import ContextData
from unikube.storage.user import LocalStorageUser, get_local_storage_user
from unikube.unikubefile.selector import unikube_file_selector
from unikube.unikubefile.unikube_file import UnikubeFile


class ContextError(Exception):
    pass


class IContext(ABC):
    @abstractmethod
    def get(self, **kwargs) -> ContextData:
        raise NotImplementedError


class ClickOptionContext(IContext):
    def __init__(self, click_options):
        self.click_options = click_options

    def get(self, **kwargs) -> ContextData:
        def _get_and_validate_argument_id(argument_name: str):
            argument = self.click_options.get(argument_name, None)
            if argument:
                if is_valid_uuid4(argument):
                    argument_id = argument
                else:
                    raise ContextError(f"Invalid {argument_name} id.")
            else:
                argument_id = None

            return argument_id

        # arguments
        organization_id = _get_and_validate_argument_id("organization")
        project_id = _get_and_validate_argument_id("project")
        deck_id = _get_and_validate_argument_id("deck")

        return ContextData(
            organization_id=organization_id,
            project_id=project_id,
            deck_id=deck_id,
        )


class UnikubeFileContext(IContext):
    def __init__(self, path_unikube_file: Union[UnikubeFile, None]):
        self.path_unikube_file = path_unikube_file

    def get(self, **kwargs) -> ContextData:
        # get unikube file
        unikube_file = unikube_file_selector.get(path_unikube_file=self.path_unikube_file)

        # check if unikube file was loaded
        if not unikube_file:
            return ContextData()

        return unikube_file.get_context()


class LocalContext(IContext):
    def __init__(self, local_storage_user: Union[LocalStorageUser, None]):
        self.local_storage_user = local_storage_user

    def get(self, **kwargs) -> ContextData:
        if not self.local_storage_user:
            return ContextData()

        user_data = self.local_storage_user.get()
        return user_data.context


class ContextLogic:
    def __init__(self, context_order: List[IContext]):
        self.context_order = context_order

    def get(self) -> ContextData:
        context = ContextData()

        for context_object in self.context_order:
            # get context variables from current context
            try:
                context_current = context_object.get(current_context=context)
            except Exception as e:
                console.debug(e)
                context_current = ContextData()

            # update context
            context_dict = context.dict()
            for key, value in context_current.dict().items():
                if context_dict[key] is None:
                    context_dict[key] = value
            context = ContextData(**context_dict)

            # check if all context variables have already been set
            if None not in context.dict().values():
                break

        return context


class Context:
    def __init__(self, auth):
        self._auth = auth

    def get(self, **kwargs) -> ContextData:
        local_storage_user = get_local_storage_user()

        context_logic = ContextLogic(
            [
                ClickOptionContext(
                    click_options={key: kwargs[key] for key in ("organization", "project", "deck") if key in kwargs}
                ),
                UnikubeFileContext(path_unikube_file="unikube.yaml"),
                LocalContext(local_storage_user=local_storage_user),
            ]
        )
        context = context_logic.get()

        # show context
        if settings.CLI_ALWAYS_SHOW_CONTEXT:
            from unikube.cli.context import show_context

            show_context(ctx=None, context=context)

        return context

    def get_context_ids_from_arguments(
        self, organization_argument: str = None, project_argument: str = None, deck_argument: str = None
    ) -> Tuple[str, str, str]:
        # convert context argments into ids
        organization_id, project_id, deck_id = convert_context_arguments(
            auth=self._auth,
            organization_argument=organization_argument,
            project_argument=project_argument,
            deck_argument=deck_argument,
        )

        # consider context
        context = self.get(organization=organization_id, project=project_id, deck=deck_id)
        return context.organization_id, context.project_id, context.deck_id



# Heavily inspired by the cleo project. Check it out.
# https://github.com/sdispater/cleo
# We use double curly braces to "escape" curly braces.
# Templates are rendered using Python's built-in `format` method for strings.


BASH_TEMPLATE = """{flag_complete_functions}
{function}()
{{
    local cur script coms opts com
    COMPREPLY=()
    _get_comp_words_by_ref -n : cur prev words
    # for an alias, get the real script behind it
    if [[ $(type -t ${{words[0]}}) == "alias" ]]; then
        script=$(alias ${{words[0]}} | sed -E "s/alias ${{words[0]}}='(.*)'/\\1/")
    else
        script=${{words[0]}}
    fi
    # lookup for command
    for word in ${{words[@]:1}}; do
        if [[ $word != -* ]]; then
            com=$word
            break
        fi
    done

    # completing for an option
    if [[ ${{cur}} == --* ]] ; then
        opts="{opts}"
        case "$com" in
{command_list}
        esac
        COMPREPLY=($(compgen -W "${{opts}}" -- ${{cur}}))
        __ltrim_colon_completions "$cur"
        return 0;
    fi

    if [[ $prev == $com ]]; then
        case "$com" in
            {subcommands}
        esac
    fi

    # completing for a command
    if [[ $cur == $com ]]; then
        coms="{coms}"
        COMPREPLY=($(compgen -W "${{coms}}" -- ${{cur}}))
        __ltrim_colon_completions "$cur"
        return 0
    fi
}}
{compdefs}"""

ZSH_TEMPLATE = """#compdef %(script_name)s
%(function)s()
{
    local state com cur
    cur=${words[${#words[@]}]}
    # lookup for command
    for word in ${words[@]:1}; do
        if [[ $word != -* ]]; then
            com=$word
            break
        fi
    done
    if [[ ${cur} == --* ]]; then
        state="option"
        opts=(%(opts)s)
    elif [[ $cur == $com ]]; then
        state="command"
        coms=(%(coms)s)
    fi
    case $state in
        (command)
            _describe 'command' coms
        ;;
        (option)
            case "$com" in
%(command_list)s
            esac
            _describe 'option' opts
        ;;
        *)
            # fallback to file completion
            _arguments '*:file:_files'
    esac
}
%(function)s "$@"
%(compdefs)s"""

FISH_TEMPLATE = """function __fish%(function)s_no_subcommand
    for i in (commandline -opc)
        if contains -- $i %(cmds_names)s
            return 1
        end
    end
    return 0
end
# global options
%(opts)s
# commands
%(cmds)s
# command options
%(cmds_opts)s"""


TEMPLATES = {"bash": BASH_TEMPLATE, "zsh": ZSH_TEMPLATE, "fish": FISH_TEMPLATE}



from tests.login_testcase import LoginTestCase
from unikube.cli import orga
from unikube.commands import ClickContext


class OrgaTestCase(LoginTestCase):
    def test_orga_info(self):
        result = self.runner.invoke(
            orga.info,
            ["ACME"],
            obj=ClickContext(),
        )

        self.assertIn("Key", result.output)
        self.assertIn("Value", result.output)
        self.assertIn("title", result.output)
        self.assertIn("ACME", result.output)
        self.assertEqual(result.exit_code, 0)

    def test_info_not_existing_orga(self):
        result = self.runner.invoke(
            orga.info,
            "not_existing_orga",
            obj=ClickContext(),
        )
        self.assertIn("[ERROR] Organization name/slug does not exist.\n", result.output)

    def test_orga_list(self):

        result = self.runner.invoke(
            orga.list,
            obj=ClickContext(),
        )

        self.assertIn("id", result.output)
        self.assertIn("name", result.output)
        self.assertIn("acme", result.output)
        self.assertEqual(result.exit_code, 0)
